"""The top-level API for typelib.

Examples: Typical Usage
    >>> import dataclasses
    >>> import typelib
    >>>
    >>> @dataclasses.dataclass
    ... class Data:
    ...     attr: int
    ...     key: str
    ...
    >>>
    >>> typelib.unmarshal(Data, '{"key":"string","attr":"1"}')
    Data(attr=1, key='string')
    >>> typelib.marshal(Data(attr=1, key='string'))
    {'attr': 1, 'key': 'string'}
    >>> codec = typelib.codec(Data)
    >>> codec.encode(Data(attr=1, key='string'))
    b'{"attr":1,"key":"string"}'
    >>> codec.decode(b'{"key":"string","attr":1}')
    Data(attr=1, key='string')
"""

from __future__ import annotations

import importlib.metadata

from typelib.api import *

__metadata__ = importlib.metadata.metadata("typelib")
__version__ = __metadata__.get("version")
__authors__ = __metadata__.get("authors")
__license__ = __metadata__.get("license")



"""Vendored class decorators for dataclasses.

Notes:
    This module is unnecessary for Python versions >= 3.10.

Info: Typical Usage
    ```pycon
    >>> import dataclasses
    >>> from typelib.py import classes
    >>>
    >>> @classes.slotted
    >>> @dataclasses.dataclass
    >>> class Slotted:
    ...     attr: str
    ...
    >>> Slotted.__slots__
    ('attr',)
    ```
"""

from __future__ import annotations

import dataclasses
import sys
import warnings
from typing import Callable, MutableSet, TypeVar, overload

from typelib import constants

_ClsT = TypeVar("_ClsT", bound=type)


@overload
def slotted(
    *, dict: bool = False, weakref: bool = True
) -> Callable[[_ClsT], _ClsT]: ...


@overload
def slotted(_cls: _ClsT, *, dict: bool = False, weakref: bool = True) -> _ClsT: ...


def slotted(  # noqa: C901
    _cls: _ClsT | None = None,
    *,
    dict: bool = False,
    weakref: bool = True,
) -> Callable[[_ClsT], _ClsT] | _ClsT:
    """Decorator to create a "slotted" version of the provided class.

    Args:
        _cls: The class to decorate.
        dict: Whether to add a slot for `__dict__`.
        weakref: Whether to add a slot for `__weakref__`.

    Warning:
        This function returns new class object as it's not possible to add `__slots__`
        after class creation.

    See Also:
        - [dataslots](https://github.com/starhel/dataslots/blob/master/src/dataslots/__init__.py)
    """

    def _slots_setstate(self, state):
        for param_dict in filter(None, state):
            for slot, value in param_dict.items():
                object.__setattr__(self, slot, value)

    def wrap(cls):
        key = repr(cls)
        if key in _stack:  # pragma: no cover
            raise TypeError(
                f"{cls!r} uses a custom metaclass {cls.__class__!r} "
                "which is not compatible with automatic slots. "
                "See Issue !typical#104 on GitHub for more information."
            ) from None

        _stack.add(key)

        if (
            sys.version_info >= (3, 10) and constants.PKG_NAME not in cls.__module__
        ):  # pragma: no cover
            warnings.warn(
                f"You are using Python {sys.version}. "
                "Python 3.10 introduced native support for slotted dataclasses. "
                "This is the preferred method for adding slots.",
                stacklevel=2,
            )

        cls_dict = {**cls.__dict__}
        # Create only missing slots
        inherited_slots = set().union(*(getattr(c, "__slots__", ()) for c in cls.mro()))

        field_names = {f.name: ... for f in dataclasses.fields(cls) if f.name}
        if dict:
            field_names["__dict__"] = ...
        if weakref:
            field_names["__weakref__"] = ...
        cls_dict["__slots__"] = (*(f for f in field_names if f not in inherited_slots),)

        # Erase filed names from class __dict__
        for f in field_names:
            cls_dict.pop(f, None)

        # Erase __dict__ and __weakref__
        cls_dict.pop("__dict__", None)
        cls_dict.pop("__weakref__", None)

        # Pickle fix for frozen dataclass as mentioned in https://bugs.python.org/issue36424
        # Use only if __getstate__ and __setstate__ are not declared and frozen=True
        if (
            all(param not in cls_dict for param in ["__getstate__", "__setstate__"])
            and cls.__dataclass_params__.frozen
        ):
            cls_dict["__setstate__"] = _slots_setstate

        # Prepare new class with slots
        new_cls = cls.__class__(cls.__name__, cls.__bases__, cls_dict)
        new_cls.__qualname__ = cls.__qualname__
        new_cls.__module__ = cls.__module__

        _stack.clear()
        return new_cls

    return wrap if _cls is None else wrap(_cls)


_stack: MutableSet[type] = set()



# flake8: noqa
from __future__ import annotations

from typing import TYPE_CHECKING

import sys

__all__ = (
    "Final",
    "ParamSpec",
    "Self",
    "TypeAliasType",
    "TypeGuard",
    "TypeIs",
    "TypeVarTuple",
    "DATACLASS_KW_ONLY",
    "DATACLASS_MATCH_ARGS",
    "DATACLASS_NATIVE_SLOTS",
    "KW_ONLY",
    "lru_cache",
    "cache",
)

if TYPE_CHECKING:
    from typing import TypeVar, overload, Callable

    DATACLASS_KW_ONLY: Final[bool] = True
    DATACLASS_NATIVE_SLOTS: Final[bool] = True
    DATACLASS_MATCH_ARGS: Final[bool] = True
    KW_ONLY: Final[object] = object()

    from typing_extensions import (
        TypeIs,
        ParamSpec,
        Self,
        Final,
        TypeAlias,
        TypeAliasType,
        TypeVarTuple,
        TypeGuard,
    )

    import orjson as json

    F = TypeVar("F")

    @overload
    def lru_cache(*, maxsize: int | None = None) -> Callable[[F], F]: ...

    @overload
    def lru_cache(func: F, *, maxsize: int | None = None) -> F: ...

    def lru_cache(
        func: F | None = None, *, maxsize: int | None = None
    ) -> F | Callable[[F], F]: ...

    def cache(func: F) -> F: ...

    if sys.version_info >= (3, 11):
        TupleVarsT = TypeVarTuple("TupleVarsT")
        TupleT: TypeAlias = "tuple[*TupleVarsT]"  # type: ignore[valid-type]

    else:
        TupleVarsT = TypeVarTuple("TupleVarsT")
        TupleT: TypeAlias = tuple[TupleVarsT]

else:
    from functools import lru_cache, cache

    if sys.version_info >= (3, 13):
        from typing import TypeIs

    else:
        from typing_extensions import TypeIs

    if sys.version_info >= (3, 12):
        from typing import TypeAliasType

    else:
        from typing_extensions import TypeAliasType

    if sys.version_info >= (3, 11):
        from typing import ParamSpec, Self, Final, TypeVarTuple

        TupleVarsT = TypeVarTuple("TupleVarsT")
        TupleT: TypeAlias = "tuple[*TupleVarsT]"

    else:
        from typing_extensions import ParamSpec, Self, Final, TypeVarTuple

        TupleVarsT = TypeVarTuple("TupleVarsT")
        TupleT: TypeAlias = tuple[TupleVarsT]

    if sys.version_info >= (3, 10):
        from typing import TypeGuard

        from dataclasses import KW_ONLY

        DATACLASS_KW_ONLY = DATACLASS_MATCH_ARGS = DATACLASS_NATIVE_SLOTS = True

    else:
        from typing_extensions import TypeGuard

        DATACLASS_KW_ONLY = DATACLASS_MATCH_ARGS = DATACLASS_NATIVE_SLOTS = False
        KW_ONLY = object()

    try:
        import orjson as json
    except (ImportError, ModuleNotFoundError):
        import json



"""Utilities for working with types as graphs.

Examples: Typical Usage
    >>> import dataclasses
    >>> from typelib import graph
    >>> graph.static_order(dict[str, str])
    [TypeNode(type=<class 'str'>, var=None, cyclic=False), TypeNode(type=dict[str, str], var=None, cyclic=False)]
    >>>
    >>> @dataclasses.dataclass
    ... class Class:
    ...     attr: str
    ...
    >>> graph.static_order(Class)
    [TypeNode(type=<class 'str'>, var='attr', cyclic=False), TypeNode(type=<class '__main__.Class'>, var=None, cyclic=False)]

"""

from __future__ import annotations

import collections
import dataclasses
import graphlib
import inspect
import typing

from typelib import constants
from typelib.py import classes, compat, inspection, refs

__all__ = ("static_order", "itertypes", "get_type_graph")


@compat.cache
def static_order(
    t: type | str | refs.ForwardRef | compat.TypeAliasType,
) -> typing.Sequence[TypeNode]:
    """Get an ordered iterable of types which resolve into the root type provided.

    Args:
        t: The type to extract an ordered stack from.

    Note:
        The order of types is guaranteed to rank from edges to root. If there are
        multiple edges, the order of those edges is not guaranteed.

        This function is memoized to avoid the cost of re-computing a type annotation
        multiple times at runtime, which would be wasted effort, as types don't change
        at runtime.

        To avoid memoization, you can make use of [`itertypes`][typelib.graph.itertypes].
    """
    # We want to leverage the cache if possible, hence the recursive call.
    #   Shouldn't actually recurse more than once or twice.
    if inspection.istypealiastype(t):
        return static_order(t.__value__)
    if isinstance(t, (str, refs.ForwardRef)):
        ref = refs.forwardref(t) if isinstance(t, str) else t
        t = refs.evaluate(ref)
        return static_order(t)

    return [*itertypes(t)]


def itertypes(
    t: type | str | refs.ForwardRef | compat.TypeAliasType,
) -> typing.Iterable[TypeNode]:
    """Iterate over the type-graph represented by `t` from edges to root.

    Args:
        t: The "root" type.

    Yields:
        [`TypeNode`][typelib.graph.TypeNode]

    Note:
        We will build a graph of types with the given type `t` as the root node,
        then iterate from the outermost leaves back to the root using BFS.

        This is computationally expensive, so you are encouraged to use
        [`static_order`][typelib.graph.static_order] instead of
        [`itertypes`][typelib.graph.itertypes].
    """
    if inspection.istypealiastype(t):
        t = t.__value__
    if isinstance(t, (str, refs.ForwardRef)):  # pragma: no cover
        ref = refs.forwardref(t) if isinstance(t, str) else t
        t = refs.evaluate(ref)

    graph = get_type_graph(t)  # type: ignore[arg-type]
    yield from graph.static_order()


def get_type_graph(t: type) -> graphlib.TopologicalSorter[TypeNode]:
    """Get a directed graph of the type(s) this annotation represents,

    Args:
        t: A type annotation.

    Returns:
        [`graphlib.TopologicalSorter`][]

    Note:
        A key aspect of building a directed graph of a given type is pre-emptive
        detection and termination of cycles in the graph. If we detect a cycle, we
        will wrap the type in a [`typing.ForwardRef`][] and mark the
        [`TypeNode`][typelib.graph.TypeNode] instance as `cyclic=True`.

        Consumers of the graph can "delay" the resolution of a forward reference until
        the graph's `static_order()` has been exhausted, at which point they have
        enough type information to resolve into the real type. (At least one layer down).

        Resolution of cyclic/recursive types is always (necessarily) lazy and should only
        resolve one level deep on each attempt, otherwise we will find ourselves stuck
        in a closed loop which never terminates (infinite recursion).
    """
    graph: graphlib.TopologicalSorter = graphlib.TopologicalSorter()
    root = TypeNode(t)
    stack = collections.deque([root])
    visited = {root.type}
    while stack:
        parent = stack.popleft()
        if inspection.isliteral(parent.type):
            graph.add(parent)
            continue

        predecessors = []
        for var, child in _level(parent.type):
            # If no type was provided, there's no reason to do further processing.
            if child in (constants.empty, typing.Any):
                continue

            # Only subscripted generics or non-stdlib types can be cyclic.
            #   i.e., we may get `str` or `datetime` any number of times,
            #   that's not cyclic, so we can just add it to the graph.
            is_visited = child in visited
            is_subscripted = inspection.issubscriptedgeneric(child)
            is_stdlib = inspection.isstdlibtype(child)
            can_be_cyclic = is_subscripted or is_stdlib is False
            # We detected a cyclic type,
            #   wrap in a ForwardRef and don't add it to the stack
            #   This will terminate this edge to prevent infinite cycles.
            if is_visited and can_be_cyclic:
                qualname = inspection.qualname(child)
                *rest, refname = qualname.split(".", maxsplit=1)
                is_argument = var is not None
                module = getattr(child, "__module__", None)
                if module in (None, "__main__") and rest:
                    module = rest[0]
                is_class = inspect.isclass(child)
                ref = refs.forwardref(
                    refname, is_argument=is_argument, module=module, is_class=is_class
                )
                node = TypeNode(ref, var=var, cyclic=True)
            # Otherwise, add the type to the stack and track that it's been seen.
            else:
                node = TypeNode(type=child, var=var)
                visited.add(node.type)
                stack.append(node)
            # Flag the type as a "predecessor" of the parent type.
            #   This lets us resolve child types first when we iterate over the graph.
            predecessors.append(node)
        # Add the parent type and its predecessors to the graph.
        graph.add(parent, *predecessors)

    return graph


@classes.slotted(dict=False, weakref=True)
@dataclasses.dataclass(unsafe_hash=True)
class TypeNode:
    """A "node" in a type graph."""

    type: typing.Any
    """The type annotation for this node."""
    var: str | None = None
    """The variable or parameter name associated to the type annotation for this node."""
    cyclic: bool = dataclasses.field(default=False, hash=False, compare=False)
    """Whether this type annotation is cyclic."""


def _level(t: typing.Any) -> typing.Iterable[tuple[str | None, type]]:
    args = inspection.args(t)
    # Only pull annotations from the signature if this is a user-defined type.
    is_structured = inspection.isstructuredtype(t)
    members = inspection.get_type_hints(t, exhaustive=is_structured)
    yield from ((None, t) for t in args)
    yield from members.items()



"""Interfaces for managing type-enforced wire protocols (codecs)."""

from __future__ import annotations

import dataclasses
import typing as t

from typelib import marshals, serdes, unmarshals
from typelib.py import classes, compat, inspection

__all__ = ("Codec", "codec")


T = t.TypeVar("T")


@compat.cache
def codec(
    t: type[T],
    *,
    marshaller: marshals.AbstractMarshaller[T] | None = None,
    unmarshaller: unmarshals.AbstractUnmarshaller[T] | None = None,
    encoder: EncoderT = compat.json.dumps,
    decoder: DecoderT = compat.json.loads,
    codec_cls: type[CodecT[T]] | None = None,
) -> CodecT[T]:
    """Factory function for creating a [`Codec`][typelib.Codec] instance.

    Note:
        In the simplest case, all that needs be provided is the first parameter (`t`). We will
        generate a marshaller, unmarshaller and build a codec. However, we provide ample
        means for customization:

        - You can pass in a subclass of [`Codec`][typelib.codecs.Codec] to `codec_cls`.
        - You may supply custom `marshaller` or `unmarshaller` callables - we will generate
          one using the high-level APIs from [`marshals`][typelib.marshals] and
          [`unmarshals`][typelib.unmarshals] if not supplied.
        - The `encoder` and `decoder` default to JSON, using either
          stdlib [`json`][] or [`orjson`](https://github.com/ijl/orjson){.external}
          if available. You may provide custom `encoder` and `decoder` callables, the only
          requirement is they ser/des to/from `bytes`.

        /// tip
        If you installed the `json` extra when you installed this library, then you have
        installed [`orjson`](https://github.com/ijl/orjson){.external}.
        ///

    Args:
        t: The type to create the interchange protocol for.
        marshaller: The marshaller used to marshal inputs into the associated type. (optional)
        unmarshaller: The unmarshaller used to unmarshal inputs into the associated type. (optional)
        encoder: The encoder for encoding data for over-the-wire (defaults to JSON).
        decoder: The decoder for decoding data from over-the-wire (defaults to JSON).
        codec_cls: The codec class definition, if overloading (optional).

    """
    marshal = marshaller or marshals.marshaller(t=t)
    unmarshal = unmarshaller or unmarshals.unmarshaller(t=t)
    cls = codec_cls or Codec
    if inspection.isbytestype(t):
        cdc = cls(
            marshal=marshal,
            unmarshal=unmarshal,
            encoder=lambda v: v,  # type: ignore[arg-type,return-value]
            decoder=lambda v: v,  # type: ignore[arg-type,return-value]
        )
        return cdc
    cdc = cls(
        marshal=marshal,
        unmarshal=unmarshal,
        encoder=encoder,
        decoder=decoder,
    )
    return cdc


@classes.slotted(dict=False, weakref=False)
@dataclasses.dataclass(frozen=True)
class Codec(t.Generic[T]):
    """A standard wire protocol (codec).

    This codec enables you to directly encode and decode your data model into your wire protocol.
    """

    marshal: marshals.AbstractMarshaller[T]
    """The marshaller used to convert an instance of `T` to a serializable object."""
    unmarshal: unmarshals.AbstractUnmarshaller[T]
    """The unmarshaller used to convert a deserialized object into an instance of `T`."""
    encoder: EncoderT
    """The encoder used to serialize a marshalled `T` into bytes."""
    decoder: DecoderT
    """The decoder used to deserialize a bytes-like object into a Python data structure for marshalling into `T`."""

    def encode(self, value: T) -> bytes:
        """Encode an instance of `T` to bytes.

        We will first marshal the given instance using the
        [`marshal`][typelib.Codec.marshal], then encode the marshalled data
        into bytes using the [`encoder`][typelib.Codec.encoder].

        Args:
            value: The instance to encode.
        """
        marshalled = self.marshal(value)
        encoded = self.encoder(marshalled)
        return encoded

    def decode(self, value: bytes) -> T:
        """Decode an instance of `T` from bytes.

        We will first decode the data from bytes using the
        [`decoder`][typelib.Codec.decoder], then unmarshal the data into an
        instance of `T` using [`unmarshal`][typelib.Codec.unmarshal].

        Args:
            value: The bytes to decode.
        """
        decoded = self.decoder(value)
        unmarshalled = self.unmarshal(decoded)
        return unmarshalled


CodecT = compat.TypeAliasType("CodecT", Codec, type_params=(T,))
"""Generic type alias with an upper bound of [`Codec`][typelib.Codec]."""
EncoderT: t.TypeAlias = t.Callable[[serdes.MarshalledValueT], bytes]
"""Protocol for a wire serializer."""
DecoderT: t.TypeAlias = t.Callable[[bytes], serdes.MarshalledValueT]
"""Protocol for a wire deserializer."""



"""Components for Python compatibility, introspection, and reflection.

Notes:
    The functionality here enables us to support multiple versions of Python to maintain
    forwards- and backwards-compatibility. This enables developers to leverage the
    bleeding edge of Python typing as new features and innovations are introduced,
    without having to upgrade to a new version of the language.

    Think of this package as a [polyfill][polyfill]{.external} for Python typing.

    [polyfill]: https://en.wikipedia.org/wiki/Polyfill_(programming)#:~:text=In%20software%20development%2C%20a%20polyfill,not%20natively%20support%20the%20feature.
"""



"""Utilities for maintaining runtime compatibility with emerging type annotation operations.

Notes:
    This module's functionality is unnecessary for Python versions >= 3.10

Examples: Typical Usage
    >>> from typelib.py import future
    >>> future.transform_annotation("str | int")
    'typing.Union[str, int]'
    >>> future.transform_annotation("dict[str, int]")
    'typing.Dict[str, int]'
"""

from __future__ import annotations

import ast
import collections
import functools
import sys
import typing

__all__ = ("transform_annotation",)


@functools.cache
def transform(annotation: str, *, union: str = "typing.Union") -> str:
    """Transform a modern annotations into their [`typing`][] equivalent:

    - [`types.UnionType`][] into a [`typing.Union`][] (`str | int` -> `typing.Union[str, int]`)
    - builtin generics into typing generics (`dict[str, int]` -> `typing.Dict[str, int]`)

    Args:
        annotation: The annotation to transform, as a string.
        union: The name of the Union type to subscript (defaults `"typing.Union"`).

    Note:
        While this transformation requires your expression be valid Python syntax, it
        doesn't make sure the type annotation is valid.
    """
    parsed = ast.parse(annotation, mode="eval")
    transformed = TransformAnnotation(union=union).generic_visit(parsed)
    unparsed = ast.unparse(transformed).strip()
    return unparsed


class TransformAnnotation(ast.NodeTransformer):
    """A [`ast.NodeTransformer`][] that transforms [`typing.Union`][]."""

    def __init__(self, union: str = "typing.Union") -> None:
        self.union = union

    def visit_BinOp(self, node: ast.BinOp):
        """Transform a [`ast.BinOp`][] to [`typing.Union`][]."""
        # Ignore anything but a bitwise OR `|`
        if not isinstance(node.op, ast.BitOr):
            return node
        # Build a stack of args to the bitor
        args = collections.deque([node.right])
        left = node.left
        while isinstance(left, ast.BinOp):
            args.appendleft(left.right)
            left = left.left
        args.appendleft(left)
        # Visit each node in the stack
        elts = [self.visit(n) for n in args]
        # Write the old-style `Union`.
        union = ast.Subscript(
            value=ast.Name(id=self.union, ctx=ast.Load()),
            slice=ast.Index(value=ast.Tuple(elts=elts, ctx=ast.Load())),  # type: ignore[call-arg,arg-type]
            ctx=ast.Load(),
        )
        ast.copy_location(union, node)
        ast.fix_missing_locations(union)
        return union

    def visit_Name(self, node: ast.Name):
        """Transform a builtin [`ast.Name`][] to the `typing` equivalent."""
        # Re-write new-style builtin generics as old-style typing generics
        if node.id not in _GENERICS:
            return node

        new = ast.Name(id=_GENERICS[node.id], ctx=ast.Load())
        ast.copy_location(new, node)
        return new

    def visit_Subscript(self, node: ast.Subscript):
        """Transform all subscripts within a [`ast.Subscript`][]."""
        # Scan all subscripts to we transform nested new-style types.
        transformed = self.visit(node.slice)
        new = ast.Subscript(
            value=self.visit(node.value),
            slice=transformed,
            ctx=node.ctx,
        )
        ast.copy_location(new, node)
        ast.fix_missing_locations(new)
        return new

    def visit_Tuple(self, node: ast.Tuple):
        """Transform all values within a [`ast.Tuple`][]."""
        # Scan all tuples to ensure we transform nested new-style types.
        transformed = [self.visit(n) for n in node.elts]
        new = ast.Tuple(elts=transformed, ctx=node.ctx)
        ast.copy_location(new, node)
        ast.fix_missing_locations(new)
        return new


_GENERICS = {
    "dict": "typing.Dict",
    "list": "typing.List",
    "set": "typing.Set",
    "tuple": "typing.Tuple",
    "Pattern": "typing.Pattern",
}


if typing.TYPE_CHECKING:  # pragma: no cover

    def transform_annotation(annotation: str, *, union: str = "Union") -> str: ...

elif sys.version_info >= (3, 10):  # pragma: no cover

    def transform_annotation(annotation: str, *, union: str = "Union") -> str:
        return annotation

else:  # pragma: no cover
    transform_annotation = transform



"""Support for unmarshalling unstructured data into Python data structures.

Notes:
    "Unmarshalling" refers to the process of taking a "primitive" form of data, such
    as a basic dictionary or JSON string, and coercing it into a higher-order structured
    data type.

Tip:
    You may use this package directly, but we encourage you to work with the
    higher-level API provided by the [`typelib`][] module.

Examples: Typical Usage
    >>> import dataclasses
    >>> import decimal
    >>> from typelib import unmarshals
    >>>
    >>> @dataclasses.dataclass(slots=True, weakref_slot=True, kw_only=True)
    ... class Struct:
    ...     key: str
    ...     number: decimal.Decimal
    ...
    >>>
    >>> data = {"key": "some-key", "number": "3.14"}
    >>> unmarshals.unmarshal(Struct, data)
    Struct(key='some-key', number=decimal.Decimal('3.14'))
    >>> unmarshaller = unmarshals.unmarshaller(Struct)
    >>> unmarshaller(data)
    Struct(key='some-key', number=decimal.Decimal('3.14'))

See Also:
    * [`unmarshals`][typelib.unmarshals.unmarshal]
    * [`unmarshaller`][typelib.unmarshals.unmarshaller]
    * [`typelib.codec`][]
"""

from typelib.unmarshals.api import *
from typelib.unmarshals.routines import *



from __future__ import annotations

import typing as t

from typelib.codecs import Codec, DecoderT, EncoderT, codec
from typelib.marshals import AbstractMarshaller, marshal, marshaller
from typelib.py import compat, refs
from typelib.unmarshals import AbstractUnmarshaller, unmarshal, unmarshaller

__all__ = (
    "AbstractMarshaller",
    "AbstractUnmarshaller",
    "Codec",
    "decode",
    "encode",
    "marshal",
    "marshaller",
    "unmarshal",
    "unmarshaller",
    "compat",
    "refs",
    "codec",
)


T = t.TypeVar("T")


def encode(
    value: T,
    *,
    t: type[T] | refs.ForwardRef | str | None = None,
    encoder: EncoderT = compat.json.dumps,
) -> bytes:
    """Encode a value into a bytes object.

    Args:
        value: The value to encode.
        t: A type hint for resolving the marshaller.
        encoder: A callable that takes a value and returns a bytes object.
    """
    marshalled = marshal(value=value, t=t)
    encoded = encoder(marshalled)
    return encoded


def decode(
    t: type[T] | refs.ForwardRef | str,
    value: bytes,
    *,
    decoder: DecoderT = compat.json.loads,
) -> T:
    """Decode a bytes object into an instance of `t`.

    Args:
        t: A type hint for resolving the unmarshaller.
        value: The value to decode.
        decoder: A callable that takes a bytes object and returns a Python value.
    """
    decoded = decoder(value)
    unmarshalled = unmarshal(t=t, value=decoded)
    return unmarshalled
